#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains methods for looking up and using CoolProp parameters
within the IDAES generic properties framework.

The CoolPropWrapper class contains a set of sub-classes for the supported
thermophysical properties required by the generic property framework, along
with some helper functions for common calls to the CoolProp database and
caching data to avoid repeated database lookups.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

import json

from pyomo.common.dependencies import attempt_import
from pyomo.environ import units as pyunits, Var

import idaes.models.properties.modular_properties.coolprop.coolprop_forms as cforms
from idaes.core.util.exceptions import BurntToast

CoolProp, coolprop_available = attempt_import("CoolProp.CoolProp")


# Map IDAES names to Coolprop names and where to look for them
name_map = {
    "dens_mol_crit": ("rhomolar", "CRITICAL"),
    "enth_mol_crit": ("hmolar", "CRITICAL"),
    "entr_mol_crit": ("smolar", "CRITICAL"),
    "mw": ("molar_mass", "EOS"),
    "omega": ("acentric", "EOS"),
    "pressure_crit": ("p", "CRITICAL"),
    "temperature_crit": ("T", "CRITICAL"),
}


class CoolPropExpressionError(ValueError):
    """
    Error message for when an unexpected expression form is called for.

    This is mostly to future proof the code against changes in CoolProp where
    the expression form is changed to something we don't recognize.
    """

    def __init__(self, prop, comp):
        msg = (
            f"Found unsupported expression form for {prop} "
            f"of component {comp}. This likely occurred due to "
            f"changes in CoolProp and the interface should be "
            f"updated."
        )

        super().__init__(self, msg)


class CoolPropPropertyError(KeyError):
    """
    Error message for when a parameter is called for that is not in CoolProp's
    database.
    """

    def __init__(self, prop, comp):
        msg = (
            f"Could not retrieve parameters for {prop} of "
            f"component {comp} from CoolProp. This likely indicates "
            f"that CoolProp does not have values for the necessary "
            f"parameters."
        )

        super().__init__(self, msg)


class CoolPropWrapper:
    """
    Interface wrapper for calling CoolProp parameter database from within the
    IDAES Generic Property Package Framework.

    This class is intended to be used in theromphysical property definition
    dicts and directs the properties framework to use forms and parameter data
    from the CoolProp libraries (if available). This requires that the user
    have CoolProp installed locally.
    """

    _cached_components = {}

    @staticmethod
    def get_parameter_value(comp_name, param):
        """
        Retrieve tuples of (value, units) for the specified property and
        component for the critical state section of the CoolProp json format.

        Args:
            comp_name: name of component for which to retirve parameters
            param: name of parameter to return value and units

        Returns:
            tuple of parameter (value, units)
        """
        map_tuple = name_map.get(param, None)
        if map_tuple is None:
            raise BurntToast(
                "Unrecognized property name in CoolProp wrapper. Please "
                "contact the IDAES developers with this bug."
            )

        if map_tuple[1] == "CRITICAL":
            ptuple = CoolPropWrapper._get_critical_property(comp_name, map_tuple[0])
        elif map_tuple[1] == "EOS":
            ptuple = CoolPropWrapper._get_eos_property(comp_name, map_tuple[0])
        else:
            raise BurntToast(
                "Unrecognized argument in CoolProp wrapper. Please contact "
                "the IDAES developers with this bug."
            )

        return ptuple

    @staticmethod
    def flush_cached_components():
        """
        Clear cached component parameter data. This sets _cached_components to
        an empty dict.

        Args:
            None

        Returns:
            None
        """
        CoolPropWrapper._cached_components = {}

    # -------------------------------------------------------------------------
    # Pure component property sub-classes
    class dens_mol_liq_comp:
        """
        Calculate liquid molar density using CoolProp forms and parameters.
        """

        @staticmethod
        def build_parameters(cobj):
            cname = cobj.local_name
            cdict = CoolPropWrapper._get_component_data(cname)

            # 8-Dec-21: CoolProp only uses rhoLnoexp for liquid density
            nlist, tlist = CoolPropWrapper._get_param_dicts(
                cname, cdict, "dens_mol_liq_comp", "rhoL", ["rhoLnoexp"]
            )

            cforms.parameters_nt_sum(cobj, "dens_mol_liq_comp", nlist, tlist)

        @staticmethod
        def return_expression(b, cobj, T):
            return cforms.expression_nonexponential(
                cobj, "dens_mol_liq_comp", T, cobj.dens_mol_crit
            )

    class enth_mol_liq_comp:
        """
        Calculate liquid molar enthalpy using CoolProp forms and parameters.
        """

        @staticmethod
        def build_parameters(cobj):
            cname = cobj.local_name
            cdict = CoolPropWrapper._get_component_data(cname)

            # 29-Dec-21: CoolProp only uses rational_polynomial.
            Alist, Blist = CoolPropWrapper._get_param_dicts(
                cname, cdict, "enth_mol_liq_comp", "hL", ["rational_polynomial"]
            )
            href = cdict["EOS"][0]["STATES"]["hs_anchor"]["hmolar"]

            cforms.parameters_polynomial(
                cobj, "enth_mol_liq_comp", pyunits.J / pyunits.mol, Alist, Blist
            )

            href_var = Var(
                doc="Reference heat of formation", units=pyunits.J / pyunits.mol
            )
            cobj.add_component("enth_mol_liq_comp_anchor", href_var)
            href_var.fix(href)

        @staticmethod
        def return_expression(b, cobj, T):
            h = (
                cforms.expression_polynomial(cobj, "enth_mol_liq_comp", T)
                + cobj.enth_mol_liq_comp_anchor
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(h, units.ENERGY_MOLE)

    class enth_mol_ig_comp:
        """
        Calculate ideal gas molar enthalpy using CoolProp forms and parameters.
        """

        @staticmethod
        def build_parameters(cobj):
            cname = cobj.local_name
            cdict = CoolPropWrapper._get_component_data(cname)

            # 29-Dec-21: CoolProp only uses rational_polynomial.
            Alist, Blist = CoolPropWrapper._get_param_dicts(
                cname, cdict, "enth_mol_ig_comp", "hLV", ["rational_polynomial"]
            )

            cforms.parameters_polynomial(
                cobj, "enth_mol_ig_comp", pyunits.J / pyunits.mol, Alist, Blist
            )

            # Next, build parameters for enth_mol_liq_comp if necessary
            if not hasattr(cobj, "enth_mol_liq_comp_coeff_A0"):
                CoolPropWrapper.enth_mol_liq_comp.build_parameters(cobj)

        @staticmethod
        def return_expression(b, cobj, T):
            h = cforms.expression_polynomial(
                cobj, "enth_mol_ig_comp", T
            ) + CoolPropWrapper.enth_mol_liq_comp.return_expression(b, cobj, T)

            units = b.params.get_metadata().derived_units
            return pyunits.convert(h, units.ENERGY_MOLE)

    class entr_mol_liq_comp:
        """
        Calculate liquid molar entropy using CoolProp forms and parameters.
        """

        @staticmethod
        def build_parameters(cobj):
            cname = cobj.local_name
            cdict = CoolPropWrapper._get_component_data(cname)

            # 29-Dec-21: CoolProp only uses rational_polynomial.
            Alist, Blist = CoolPropWrapper._get_param_dicts(
                cname, cdict, "entr_mol_liq_comp", "sL", ["rational_polynomial"]
            )
            sref = cdict["EOS"][0]["STATES"]["hs_anchor"]["smolar"]

            cforms.parameters_polynomial(
                cobj,
                "entr_mol_liq_comp",
                pyunits.J / pyunits.mol / pyunits.K,
                Alist,
                Blist,
            )

            sref_var = Var(
                doc="Reference heat of formation",
                units=pyunits.J / pyunits.mol / pyunits.K,
            )
            cobj.add_component("entr_mol_liq_comp_anchor", sref_var)
            sref_var.fix(sref)

        @staticmethod
        def return_expression(b, cobj, T):
            s = (
                cforms.expression_polynomial(cobj, "entr_mol_liq_comp", T)
                + cobj.entr_mol_liq_comp_anchor
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(s, units.ENTROPY_MOLE)

    class entr_mol_ig_comp:
        """
        Calculate ideal gas molar entropy using CoolProp forms and parameters.
        """

        @staticmethod
        def build_parameters(cobj):
            cname = cobj.local_name
            cdict = CoolPropWrapper._get_component_data(cname)

            # 29-Dec-21: CoolProp only uses rational_polynomial.
            Alist, Blist = CoolPropWrapper._get_param_dicts(
                cname, cdict, "entr_mol_ig_comp", "sLV", ["rational_polynomial"]
            )

            cforms.parameters_polynomial(
                cobj,
                "entr_mol_ig_comp",
                pyunits.J / pyunits.mol / pyunits.K,
                Alist,
                Blist,
            )

            # Next, build parameters for entr_mol_liq_comp if necessary
            if not hasattr(cobj, "entr_mol_liq_comp_coeff_A0"):
                CoolPropWrapper.entr_mol_liq_comp.build_parameters(cobj)

        @staticmethod
        def return_expression(b, cobj, T):
            s = cforms.expression_polynomial(
                cobj, "entr_mol_ig_comp", T
            ) + CoolPropWrapper.entr_mol_liq_comp.return_expression(b, cobj, T)

            units = b.params.get_metadata().derived_units
            return pyunits.convert(s, units.ENTROPY_MOLE)

    class pressure_sat_comp:
        """
        Calculate pure component saturation pressure using CoolProp forms and
        parameters.
        """

        @staticmethod
        def build_parameters(cobj):
            cname = cobj.local_name
            cdict = CoolPropWrapper._get_component_data(cname)

            # 8-Dec-21: CoolProp only has two "types" for pressure_sat
            # which appear to be equivalent.
            nlist, tlist = CoolPropWrapper._get_param_dicts(
                cname, cdict, "pressure_sat", "pS", ["pL", "pV"], using_tau_r=True
            )

            cforms.parameters_nt_sum(cobj, "pressure_sat", nlist, tlist)

        @staticmethod
        def return_expression(b, cobj, T, dT=False):
            if dT:
                return CoolPropWrapper.pressure_sat_comp.dT_expression(b, cobj, T)

            return cforms.expression_exponential(
                cobj, "pressure_sat", T, cobj.pressure_crit, tau=True
            )

        @staticmethod
        def dT_expression(b, cobj, T):
            return cforms.dT_expression_exponential(
                cobj, "pressure_sat", T, cobj.pressure_crit, tau=True
            )

    # -------------------------------------------------------------------------
    # Internal methods

    @staticmethod
    def _get_component_data(comp_name):
        """
        Method to get parameter data from cached components, and to call
        _load_component_data if it is not present. This method also includes
        checks to handle aliases of components.

        Args:
            comp_name: name of component to get data for

        Returns:
            dict constructed from json string retrieved from CoolProp database.
        """
        if comp_name in CoolPropWrapper._cached_components:
            # First check to see if component present by comp_name
            return CoolPropWrapper._cached_components[comp_name]
        else:
            # Check to see if comp_name is an alias for a cached component
            for v in CoolPropWrapper._cached_components.values():
                if comp_name in v["INFO"]["ALIASES"] or comp_name in v["INFO"]["NAME"]:
                    CoolPropWrapper._cached_components[comp_name] = v
                    return v

        # If we haven't returned yet, then we need to load the component
        return CoolPropWrapper._load_component_data(comp_name)

    @staticmethod
    def _load_component_data(comp_name):
        """
        Method to load parameter data for specified component from the CoolProp
        database in json format. Loaded data is in dict form and is stored in
        _cached_components to avoid need for repeated calls to CoolProp.

        Args:
            comp_name: name of component to retrieve parameters for.

        Returns:
            dict constructed from json string retrieved from CoolProp database.

        Raises:
            RuntimeError is component is not found in database
        """
        try:
            prop_str = CoolProp.get_fluid_param_string(comp_name, "JSON")
        except RuntimeError:
            raise RuntimeError(
                f"Failed to find component {comp_name} in CoolProp JSON " f"database."
            )
        comp_prop = json.loads(prop_str)[0]

        CoolPropWrapper._cached_components[comp_name] = comp_prop

        return comp_prop

    @staticmethod
    def _get_critical_property(comp_name, prop_name):
        """
        Method to retrieve tuples of (value, units) for the specified property
        and component for the critical state section of the CoolProp json
        format.
        """
        cdict = CoolPropWrapper._get_component_data(comp_name)

        pc = cdict["STATES"]["critical"][prop_name]
        punits = getattr(pyunits, cdict["STATES"]["critical"][prop_name + "_units"])

        return (pc, punits)

    @staticmethod
    def _get_eos_property(comp_name, prop_name):
        """
        Method to retrieve tuples of (value, units) for the specified property
        and component for the EoS section of the CoolProp json format.
        """
        cdict = CoolPropWrapper._get_component_data(comp_name)

        pc = cdict["EOS"][0][prop_name]
        punits = cdict["EOS"][0][prop_name + "_units"]
        if punits == "-":
            punits = pyunits.dimensionless
        else:
            punits = getattr(pyunits, punits)

        return (pc, punits)

    @staticmethod
    def _get_param_dicts(
        comp_name,
        comp_data,
        prop_name,
        coolprop_name,
        expected_forms,
        using_tau_r=False,
    ):
        """
        Method to get parameter sets for expression forms. Also includes check
        to verify the expression form listed by CoolProp matches the expected
        for in IDAES.

        Args:
            comp_name: name of the component for which parameters are required
            comp_data: dict of parameter values for component (from CoolProp)
            prop_name: IDAES name of property for which parameters are required
            coolprop_name: name used by CoolProp for property of interest
            expected_forms: list of expression forms supported by wrapper for
                              for property
            using_tau_r: (optional) flag indicating whether to check for use
                          of tau_r in the epxression or not (default = False,
                          do not check).

        Returns:
            tuple of lists of parameters required by expression
        """
        try:
            # First, check to make sure the listed expression form is
            # supported.
            if comp_data["ANCILLARIES"][coolprop_name]["type"] not in expected_forms:
                # If not one of the forms we recognise, raise an exception
                raise CoolPropExpressionError(prop_name, comp_name)
            elif (
                using_tau_r
                and not comp_data["ANCILLARIES"][coolprop_name]["using_tau_r"]
            ):
                # If not one of the forms we recognise, raise an exception
                raise CoolPropExpressionError(prop_name, comp_name)

            if expected_forms == ["rational_polynomial"]:
                list1 = comp_data["ANCILLARIES"][coolprop_name]["A"]
                list2 = comp_data["ANCILLARIES"][coolprop_name]["B"]
            else:
                list1 = comp_data["ANCILLARIES"][coolprop_name]["n"]
                list2 = comp_data["ANCILLARIES"][coolprop_name]["t"]
        except KeyError:
            raise CoolPropPropertyError(prop_name, comp_name)

        return (list1, list2)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example property package for the VLE calucations for a Benzene-Toluene-o-Xylene
system. If using the activity coefficient models (NRTL or Wilson), the user is
expected to provide the parameters necessary for these models. Please note that
these parameters are declared as variables here to allow for use in a parameter
estimation problem if the VLE data is available.
"""

# Import Pyomo libraries
from pyomo.environ import Param, NonNegativeReals, Set, units as pyunits

# Import IDAES cores
from idaes.core import declare_process_block_class, Component
from idaes.core.util.misc import extract_data

from idaes.models.properties.activity_coeff_models.activity_coeff_prop_pack import (
    ActivityCoeffParameterData,
)
from idaes.logger import getIdaesLogger

# Some more information about this module
__author__ = "Jaffer Ghouse"
__version__ = "0.0.1"


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("BTXParameterBlock")
class BTXParameterData(ActivityCoeffParameterData):
    """Property package for mixtures of benzene, toluene and (ortho-)xylene."""

    def build(self):
        """
        Callable method for Block construction.
        """
        self.component_list_master = Set(initialize=["benzene", "toluene", "o-xylene"])

        # Create component objects
        # NOTE: User needs to update this list; can be a subset or
        # equal to the master component list
        self.benzene = Component()
        self.toluene = Component()

        super(BTXParameterData, self).build()

        # List of phase equilibrium index
        self.phase_equilibrium_idx_master = Set(initialize=[1, 2, 3])

        self.phase_equilibrium_idx = Set(initialize=[1, 2])

        self.phase_equilibrium_list_master = {
            1: ["benzene", ("Vap", "Liq")],
            2: ["toluene", ("Vap", "Liq")],
            3: ["o-xylene", ("Vap", "Liq")],
        }

        self.phase_equilibrium_list = {
            1: ["benzene", ("Vap", "Liq")],
            2: ["toluene", ("Vap", "Liq")],
        }

        # Thermodynamic reference state
        self.pressure_reference = Param(
            mutable=True,
            default=101325,
            doc="Reference pressure [Pa]",
            units=pyunits.Pa,
        )
        self.temperature_reference = Param(
            mutable=True,
            default=298.15,
            doc="Reference temperature [K]",
            units=pyunits.K,
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        pressure_critical_data = {
            "benzene": 48.9e5,
            "toluene": 41e5,
            "o-xylene": 37.3e5,
        }

        self.pressure_critical = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=True,
            initialize=extract_data(pressure_critical_data),
            doc="Critical pressure [Pa]",
            units=pyunits.Pa,
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        temperature_critical_data = {
            "benzene": 562.2,
            "toluene": 591.8,
            "o-xylene": 630.3,
        }

        self.temperature_critical = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=True,
            initialize=extract_data(temperature_critical_data),
            doc="Critical temperature [K]",
            units=pyunits.K,
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        mw_comp_data = {
            "benzene": 78.1136e-3,
            "toluene": 92.1405e-3,
            "o-xylene": 106.167e-3,
        }

        self.mw_comp = Param(
            self.component_list,
            mutable=True,
            initialize=extract_data(mw_comp_data),
            doc="molecular weight kg/mol",
            units=pyunits.kg / pyunits.mol,
        )

        # Constants for specific heat capacity, enthalpy
        # Sources: The Properties of Gases and Liquids (1987)
        #         4th edition, Chemical Engineering Series - Robert C. Reid
        #         Perry's Chemical Engineers Handbook
        #         - Robert H. Perry (Cp_liq)
        Cp_Liq_A_data = {("benzene"): 1.29e5, ("toluene"): 1.40e5, ("o-xylene"): 3.65e4}
        Cp_Liq_B_data = {
            ("benzene"): -1.7e2,
            ("toluene"): -1.52e2,
            ("o-xylene"): 1.0175e3,
        }
        Cp_Liq_C_data = {
            ("benzene"): 6.48e-1,
            ("toluene"): 6.95e-1,
            ("o-xylene"): -2.63,
        }
        Cp_Liq_D_data = {("benzene"): 0, ("toluene"): 0, ("o-xylene"): 3.02e-3}
        Cp_Liq_E_data = {("benzene"): 0, ("toluene"): 0, ("o-xylene"): 0}

        self.cp_mol_liq_comp_coeff_A = Param(
            self.component_list,
            units=pyunits.J / pyunits.kmol / pyunits.K,
            doc="Liquid phase Cp parameter A",
            initialize=extract_data(Cp_Liq_A_data),
        )

        self.cp_mol_liq_comp_coeff_B = Param(
            self.component_list,
            units=pyunits.J / pyunits.kmol / pyunits.K**2,
            doc="Liquid phase Cp parameter B",
            initialize=extract_data(Cp_Liq_B_data),
        )

        self.cp_mol_liq_comp_coeff_C = Param(
            self.component_list,
            units=pyunits.J / pyunits.kmol / pyunits.K**3,
            doc="Liquid phase Cp parameter C",
            initialize=extract_data(Cp_Liq_C_data),
        )

        self.cp_mol_liq_comp_coeff_D = Param(
            self.component_list,
            units=pyunits.J / pyunits.kmol / pyunits.K**4,
            doc="Liquid phase Cp parameter D",
            initialize=extract_data(Cp_Liq_D_data),
        )

        self.cp_mol_liq_comp_coeff_E = Param(
            self.component_list,
            units=pyunits.J / pyunits.kmol / pyunits.K**5,
            doc="Liquid phase Cp parameter E",
            initialize=extract_data(Cp_Liq_E_data),
        )

        Cp_Vap_A_data = {
            ("benzene"): -3.392e1,
            ("toluene"): -2.435e1,
            ("o-xylene"): -1.585e-1,
        }
        Cp_Vap_B_data = {
            ("benzene"): 4.739e-1,
            ("toluene"): 5.125e-1,
            ("o-xylene"): 5.962e-1,
        }
        Cp_Vap_C_data = {
            ("benzene"): -3.017e-4,
            ("toluene"): -2.765e-4,
            ("o-xylene"): -3.443e-4,
        }
        Cp_Vap_D_data = {
            ("benzene"): 7.130e-8,
            ("toluene"): 4.911e-8,
            ("o-xylene"): 7.528e-8,
        }
        Cp_Vap_E_data = {("benzene"): 0, ("toluene"): 0, ("o-xylene"): 0}

        self.cp_mol_vap_comp_coeff_A = Param(
            self.component_list,
            units=pyunits.J / pyunits.mol / pyunits.K,
            doc="Vapor phase Cp parameter A",
            initialize=extract_data(Cp_Vap_A_data),
        )

        self.cp_mol_vap_comp_coeff_B = Param(
            self.component_list,
            units=pyunits.J / pyunits.mol / pyunits.K**2,
            doc="Vapor phase Cp parameter B",
            initialize=extract_data(Cp_Vap_B_data),
        )

        self.cp_mol_vap_comp_coeff_C = Param(
            self.component_list,
            units=pyunits.J / pyunits.mol / pyunits.K**3,
            doc="Vapor phase Cp parameter C",
            initialize=extract_data(Cp_Vap_C_data),
        )

        self.cp_mol_vap_comp_coeff_D = Param(
            self.component_list,
            units=pyunits.J / pyunits.mol / pyunits.K**4,
            doc="Vapor phase Cp parameter D",
            initialize=extract_data(Cp_Vap_D_data),
        )

        self.cp_mol_vap_comp_coeff_E = Param(
            self.component_list,
            units=pyunits.J / pyunits.mol / pyunits.K**5,
            doc="Vapor phase Cp parameter E",
            initialize=extract_data(Cp_Vap_E_data),
        )

        # Source: The Properties of Gases and Liquids (1987)
        # 4th edition, Chemical Engineering Series - Robert C. Reid
        pressure_sat_coeff_data = {
            ("benzene", "A"): -6.98273,
            ("benzene", "B"): 1.33213,
            ("benzene", "C"): -2.62863,
            ("benzene", "D"): -3.33399,
            ("toluene", "A"): -7.28607,
            ("toluene", "B"): 1.38091,
            ("toluene", "C"): -2.83433,
            ("toluene", "D"): -2.79168,
            ("o-xylene", "A"): -7.53357,
            ("o-xylene", "B"): 1.40968,
            ("o-xylene", "C"): -3.10985,
            ("o-xylene", "D"): -2.85992,
        }

        self.pressure_sat_coeff = Param(
            self.component_list,
            ["A", "B", "C", "D"],
            mutable=False,
            initialize=extract_data(pressure_sat_coeff_data),
            doc="parameters to compute P_sat",
        )

        # Standard heats of formation
        # Source: NIST Webbook, https://webbook.nist.gov
        # Retrieved 25th September 2019
        dh_form_data = {
            ("Vap", "benzene"): 82.9e3,
            ("Vap", "toluene"): 50.1e3,
            ("Vap", "o-xylene"): 19.0e3,
            ("Liq", "benzene"): 49.0e3,
            ("Liq", "toluene"): 12.0e3,
            ("Liq", "o-xylene"): -24.4e3,
        }

        self.dh_form = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize=extract_data(dh_form_data),
            doc="Standard heats of formation [J/mol]",
            units=pyunits.J / pyunits.mol,
        )

        # Standard entropy of formation
        # Source: Engineering Toolbox, https://www.engineeringtoolbox.com
        # o-xylene from NIST Webbook, https://webbook.nist.gov
        # Retrieved 9th October, 2019
        ds_form_data = {
            ("Vap", "benzene"): -269,
            ("Vap", "toluene"): -321,
            ("Vap", "o-xylene"): -353.6,
            ("Liq", "benzene"): -173,
            ("Liq", "toluene"): -220,
            ("Liq", "o-xylene"): -246,
        }

        self.ds_form = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize=extract_data(ds_form_data),
            doc="Standard entropy of formation [J/mol.K]",
            units=pyunits.J / pyunits.mol / pyunits.K,
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for InitializerBase class
"""
import pytest
import re
import types
import os

from pyomo.environ import Block, ConcreteModel, Constraint, value, Var

from idaes.core.base.process_base import ProcessBaseBlock
from idaes.core.initialization.initializer_base import (
    InitializerBase,
    InitializationStatus,
    ModularInitializerBase,
)
from idaes.core.initialization.block_triangularization import (
    BlockTriangularizationInitializer,
)

from idaes.core.util.exceptions import InitializationError
import idaes.logger as idaeslog

path = os.path.dirname(__file__)
fname = os.path.join(path, "init_example.json")

__author__ = "Andrew Lee"


class TestSubMethods:
    @pytest.mark.unit
    def test_init(self):
        initializer = InitializerBase()

        assert initializer.summary == {}
        assert initializer.initial_state == {}

        assert hasattr(initializer, "config")

        assert "constraint_tolerance" in initializer.config

    @pytest.fixture
    def model(self):
        m = ConcreteModel()

        m.v1 = Var()
        m.v2 = Var()

        m.c1 = Constraint(expr=m.v1 == 2)
        m.c2 = Constraint(expr=m.v1 == m.v2)
        m.c3 = Constraint(expr=m.v2 >= 0)
        m.c4 = Constraint(expr=m.v1 <= 5)

        return m

    @pytest.mark.unit
    def test_get_initial_state(self, model):
        model.v1.fix(42)
        model.v2.set_value(43)

        initializer = InitializerBase()
        state = initializer.get_current_state(model)

        assert state is initializer.initial_state[model]

        expected = {
            "__type__": "<class 'pyomo.core.base.PyomoModel.ConcreteModel'>",
            "__id__": 0,
            "active": True,
            "data": {
                "None": {
                    "__type__": "<class 'pyomo.core.base.PyomoModel.ConcreteModel'>",
                    "__id__": 1,
                    "active": True,
                    "__pyomo_components__": {
                        "v1": {
                            "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                            "__id__": 2,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                                    "__id__": 3,
                                    "fixed": True,
                                    "value": 42,
                                }
                            },
                        },
                        "v2": {
                            "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                            "__id__": 4,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                                    "__id__": 5,
                                    "fixed": False,
                                    "value": 43,
                                }
                            },
                        },
                        "c1": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 6,
                            "active": True,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 7,
                                    "active": True,
                                }
                            },
                        },
                        "c2": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 8,
                            "active": True,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 9,
                                    "active": True,
                                }
                            },
                        },
                        "c3": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 10,
                            "active": True,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 11,
                                    "active": True,
                                }
                            },
                        },
                        "c4": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 12,
                            "active": True,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 13,
                                    "active": True,
                                }
                            },
                        },
                    },
                }
            },
        }

        assert expected == state["unknown"]

    @pytest.mark.unit
    def test_get_and_restore_initial_state(self, model):
        # Set an initial state
        model.v1.fix(10)
        model.c4.deactivate()

        initializer = InitializerBase()

        # Store the initial state
        initializer.get_current_state(model)

        expected = {
            "__type__": "<class 'pyomo.core.base.PyomoModel.ConcreteModel'>",
            "__id__": 0,
            "active": True,
            "data": {
                "None": {
                    "__type__": "<class 'pyomo.core.base.PyomoModel.ConcreteModel'>",
                    "__id__": 1,
                    "active": True,
                    "__pyomo_components__": {
                        "v1": {
                            "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                            "__id__": 2,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                                    "__id__": 3,
                                    "fixed": True,
                                    "value": 10,
                                }
                            },
                        },
                        "v2": {
                            "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                            "__id__": 4,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.var.ScalarVar'>",
                                    "__id__": 5,
                                    "fixed": False,
                                    "value": None,
                                }
                            },
                        },
                        "c1": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 6,
                            "active": True,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 7,
                                    "active": True,
                                }
                            },
                        },
                        "c2": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 8,
                            "active": True,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 9,
                                    "active": True,
                                }
                            },
                        },
                        "c3": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 10,
                            "active": True,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 11,
                                    "active": True,
                                }
                            },
                        },
                        "c4": {
                            "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                            "__id__": 12,
                            "active": False,
                            "data": {
                                "None": {
                                    "__type__": "<class 'pyomo.core.base.constraint.ScalarConstraint'>",
                                    "__id__": 13,
                                    "active": False,
                                }
                            },
                        },
                    },
                }
            },
        }
        assert expected == initializer.initial_state[model]["unknown"]

        # Make some more changes to the state
        model.v1.set_value(21)
        model.v1.unfix()
        model.v2.fix(22)
        model.c1.deactivate()
        model.c4.activate()

        initializer.restore_model_state(model)

        # Check that state was reverted correctly
        assert model.v1.value == 10  # Value should have changed back to original
        assert model.v1.fixed  # Should be fixed again
        assert model.v2.value == 22  # Value should not have changed
        assert not model.v2.fixed  # Should have been unfixed
        assert model.c1.active
        assert model.c2.active
        assert model.c3.active
        assert not model.c4.active

    @pytest.mark.unit
    def test_load_value_from_file(self):
        m = ConcreteModel()

        m.v1 = Var()
        m.v2 = Var()
        m.v3 = Var()
        m.v4 = Var()

        # Fix value of v2, it should not get overwritten
        m.v2.fix(10)

        initializer = InitializerBase()

        # Load values from example file, all values are 4
        initializer.load_initial_guesses(m, json_file=fname)

        assert m.v1.value == 4
        assert m.v2.value == 10  # Should retain fixed value
        assert m.v3.value == 4
        assert m.v4.value == 4

    @pytest.mark.unit
    def test_load_values_dict(self, model):
        model.v3 = Var(["a", "b"])
        model.v4 = Var(["a", "b"])
        initializer = InitializerBase()

        init_dict = {"v1": 10, "v2": 20, "v3[a]": 30, "v3[b]": 40, "v4": 50}

        initializer.load_initial_guesses(model, initial_guesses=init_dict)

        assert model.v1.value == 10
        assert model.v2.value == 20
        assert model.v3["a"].value == 30
        assert model.v3["b"].value == 40
        assert model.v4["a"].value == 50
        assert model.v4["b"].value == 50

    @pytest.mark.unit
    def test_load_values_dict_and_file(self, model):
        initializer = InitializerBase()

        with pytest.raises(
            ValueError,
            match="Cannot provide both a set of initial guesses and a json file to load.",
        ):
            initializer.load_initial_guesses(
                model, initial_guesses="foo", json_file="bar"
            )
        assert initializer.summary[model]["status"] == InitializationStatus.Error

    @pytest.mark.unit
    def test_load_values_dict_non_var(self, model):
        initializer = InitializerBase()

        init_dict = {"c1": 10}

        with pytest.raises(
            TypeError,
            match="Component c1 is not a Var. Initial guesses should only contain values for variables.",
        ):
            initializer.load_initial_guesses(model, initial_guesses=init_dict)
        assert initializer.summary[model]["status"] == InitializationStatus.Error

    @pytest.mark.unit
    def test_load_value_none(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v1 = Var()

        initializer = InitializerBase()
        initializer.load_initial_guesses(m)

        expected = "No initial guesses provided during initialization of model unknown."
        assert expected in caplog.text

        assert m.v1.value is None

    @pytest.mark.unit
    def test_fix_states(self, model):
        # Create a dummy method to fix states on test model
        def fix_initialization_states(blk):
            blk.v1.fix(12)

        model.fix_initialization_states = types.MethodType(
            fix_initialization_states, model
        )

        initializer = InitializerBase()
        initializer.fix_initialization_states(model)

        assert model.v1.fixed
        assert not model.v2.fixed

    @pytest.mark.unit
    def test_fix_states_no_method(self, model, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        initializer = InitializerBase()
        initializer.fix_initialization_states(model)

        expected = "Model unknown does not have a fix_initialization_states method - attempting to continue."
        assert expected in caplog.text

        assert not model.v1.fixed
        assert not model.v2.fixed

    @pytest.mark.unit
    def test_initialization_routine(self, model):
        initializer = InitializerBase()

        with pytest.raises(NotImplementedError):
            initializer.initialization_routine(model)
        assert initializer.summary[model]["status"] == InitializationStatus.Error

    @pytest.mark.unit
    def test_precheck_fail(self, model):
        model.v1.fix(10)

        initializer = InitializerBase()
        with pytest.raises(
            InitializationError,
            match=re.escape(
                "Degrees of freedom for unknown were not equal to zero during "
                "initialization (DoF = -1)."
            ),
        ):
            initializer.precheck(model)
        assert initializer.summary[model]["status"] == InitializationStatus.DoF
        assert initializer.summary[model]["DoF"] == -1

    @pytest.mark.unit
    def test_precheck_ok(self, model):
        initializer = InitializerBase()
        initializer.precheck(model)

        assert initializer.summary[model]["status"] == InitializationStatus.none
        assert initializer.summary[model]["DoF"] == 0

    @pytest.mark.unit
    def test_postcheck_vars_none(self, model):
        initializer = InitializerBase()

        with pytest.raises(
            InitializationError,
            match=f"unknown failed to initialize successfully: uninitialized variables or "
            "unconverged equality constraints detected. Please check postcheck summary for "
            "more information.",
        ):
            initializer.postcheck(model)

        assert initializer.summary[model]["status"] == InitializationStatus.Failed
        assert len(initializer.summary[model]["uninitialized_vars"]) == 2
        assert len(initializer.summary[model]["unconverged_constraints"]) == 4

    @pytest.mark.unit
    def test_postcheck_vars_wrong(self, model):
        model.v1.set_value(10)
        model.v2.set_value(-10)

        initializer = InitializerBase()

        with pytest.raises(
            InitializationError,
            match=f"unknown failed to initialize successfully: uninitialized variables or "
            "unconverged equality constraints detected. Please check postcheck summary for "
            "more information.",
        ):
            initializer.postcheck(model)

        assert initializer.summary[model]["status"] == InitializationStatus.Failed
        assert len(initializer.summary[model]["uninitialized_vars"]) == 0
        assert len(initializer.summary[model]["unconverged_constraints"]) == 4

    @pytest.mark.unit
    def test_postcheck_partially_satisfied(self, model):
        model.v1.set_value(-10)
        model.v2.set_value(-10)

        initializer = InitializerBase()

        with pytest.raises(
            InitializationError,
            match=f"unknown failed to initialize successfully: uninitialized variables or "
            "unconverged equality constraints detected. Please check postcheck summary for "
            "more information.",
        ):
            initializer.postcheck(model)

        assert initializer.summary[model]["status"] == InitializationStatus.Failed
        assert len(initializer.summary[model]["uninitialized_vars"]) == 0
        assert len(initializer.summary[model]["unconverged_constraints"]) == 2

    @pytest.mark.unit
    def test_postcheck_correct(self, model):
        model.v1.set_value(2)
        model.v2.set_value(2)

        initializer = InitializerBase()
        initializer.postcheck(model)

        assert initializer.summary[model]["status"] == InitializationStatus.Ok
        assert len(initializer.summary[model]["uninitialized_vars"]) == 0
        assert len(initializer.summary[model]["unconverged_constraints"]) == 0

    @pytest.mark.unit
    def test_postcheck_near_tolerance(self, model):
        model.v1.set_value(2)
        model.v2.set_value(2.0001)

        initializer = InitializerBase()
        with pytest.raises(
            InitializationError,
            match=f"unknown failed to initialize successfully: uninitialized variables or "
            "unconverged equality constraints detected. Please check postcheck summary for "
            "more information.",
        ):
            initializer.postcheck(model)

        assert initializer.summary[model]["status"] == InitializationStatus.Failed
        assert len(initializer.summary[model]["uninitialized_vars"]) == 0
        assert len(initializer.summary[model]["unconverged_constraints"]) == 1

        # Change tolerance and recheck
        initializer.config.constraint_tolerance = 1e-3

        initializer.postcheck(model)

        assert initializer.summary[model]["status"] == InitializationStatus.Ok
        assert len(initializer.summary[model]["uninitialized_vars"]) == 0
        assert len(initializer.summary[model]["unconverged_constraints"]) == 0

    @pytest.mark.unit
    def test_postcheck_ignore_unused(self):
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var()  # unused
        m.c = Constraint(expr=m.v1 * 2 == m.v2)

        initializer = InitializerBase()

        # By default ,will fail due to m.v3
        with pytest.raises(
            InitializationError,
            match=f"unknown failed to initialize successfully: uninitialized variables or "
            "unconverged equality constraints detected. Please check postcheck summary for "
            "more information.",
        ):
            initializer.postcheck(m)

        assert initializer.summary[m]["status"] == InitializationStatus.Failed
        assert initializer.summary[m]["uninitialized_vars"] == [m.v3]
        assert len(initializer.summary[m]["unconverged_constraints"]) == 0

        # Try again whilst skipping unused vars
        initializer.postcheck(m, exclude_unused_vars=True)

        assert initializer.summary[m]["status"] == InitializationStatus.Ok
        assert len(initializer.summary[m]["uninitialized_vars"]) == 0
        assert len(initializer.summary[m]["unconverged_constraints"]) == 0

    @pytest.mark.unit
    def test_update_summary(self):
        initializer = InitializerBase()

        assert initializer.summary == {}

        initializer._update_summary("foo", "bar", "baz")

        assert initializer.summary["foo"]["bar"] == "baz"
        assert initializer.summary["foo"]["status"] == InitializationStatus.none

    @pytest.mark.unit
    def test_load_values_from_dict_not_var(self):
        m = ConcreteModel()
        m.v = Var()
        m.c = Constraint(expr=m.v == 0)

        initializer = InitializerBase()

        with pytest.raises(
            TypeError,
            match="Component c is not a Var. Initial guesses should only contain values for variables.",
        ):
            initializer._load_values_from_dict(m, {m.c: 10})

        assert initializer.summary[m]["status"] == InitializationStatus.Error

    @pytest.mark.unit
    def test_load_values_from_dict_indexed(self):
        m = ConcreteModel()
        m.v = Var(["a", "b"])

        initializer = InitializerBase()

        initializer._load_values_from_dict(m, {m.v: 10})

        assert m.v["a"].value == 10
        assert m.v["b"].value == 10

    @pytest.mark.unit
    def test_load_values_from_dict_indexed_fixed(self):
        m = ConcreteModel()
        m.v = Var(["a", "b"])
        m.v["b"].fix(20)

        initializer = InitializerBase()

        with pytest.raises(
            InitializationError,
            match=re.escape(
                "Attempted to change the value of fixed variable v[b]. "
                "Initialization from initial guesses does not support changing the value "
                "of fixed variables."
            ),
        ):
            initializer._load_values_from_dict(m, {m.v: 10})

        assert initializer.summary[m]["status"] == InitializationStatus.Error

    @pytest.mark.unit
    def test_load_values_from_dict_indexed_fixed_ignore(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var(["a", "b"])
        m.v["b"].fix(20)

        initializer = InitializerBase()
        initializer._load_values_from_dict(m, {m.v: 10}, exception_on_fixed=False)

        assert m.v["a"].value == 10
        assert m.v["b"].value == 20

        expected = "Found initial guess for fixed Var v[b] - ignoring."
        assert expected in caplog.text

    @pytest.mark.unit
    def test_load_values_from_dict_scalar(self):
        m = ConcreteModel()
        m.v = Var(["a", "b"])

        initializer = InitializerBase()

        initializer._load_values_from_dict(m, {'v["a"]': 10})

        assert m.v["a"].value == 10
        assert m.v["b"].value == None

    @pytest.mark.unit
    def test_load_values_from_dict_scalar_fixed(self):
        m = ConcreteModel()
        m.v = Var(["a", "b"])
        m.v["a"].fix(20)

        initializer = InitializerBase()

        with pytest.raises(
            InitializationError,
            match=re.escape(
                "Attempted to change the value of fixed variable v[a]. "
                "Initialization from initial guesses does not support changing the value "
                "of fixed variables."
            ),
        ):
            initializer._load_values_from_dict(m, {'v["a"]': 10})

        assert initializer.summary[m]["status"] == InitializationStatus.Error

    @pytest.mark.unit
    def test_load_values_from_dict_scalar_fixed_ignore(self, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger="idaes",
        )

        m = ConcreteModel()
        m.v = Var()
        m.v.fix(20)

        initializer = InitializerBase()
        initializer._load_values_from_dict(m, {"v": 10}, exception_on_fixed=False)

        assert m.v.value == 20

        expected = "Found initial guess for fixed Var v - ignoring."
        assert expected in caplog.text


class DummyInit:
    def plugin_prepare(self, plugin, output="default"):
        plugin._test = output

    def plugin_initialize(self, model, **kwargs):
        model._plugin_initialized = True

    def plugin_finalize(self, plugin, output="default"):
        plugin._test2 = output

    def initialize(self, model, **kwargs):
        model._initialized = True


class TestModularInitializerBase:
    @pytest.mark.unit
    def test_base_attributed(self):
        initializer = ModularInitializerBase()

        assert initializer.submodel_initializers == {}
        assert initializer.config.default_submodel_initializer is None

        assert initializer._solver is None
        assert initializer.config.solver == "ipopt_v2"
        assert initializer.config.solver_options == {}
        assert initializer.config.writer_config == {}

    @pytest.mark.unit
    def test_get_submodel_initializer_specific_model(self):
        m = ConcreteModel()

        initializer = ModularInitializerBase()
        initializer.submodel_initializers[m] = DummyInit

        assert isinstance(initializer.get_submodel_initializer(m), DummyInit)

    @pytest.mark.unit
    def test_get_submodel_initializer_model_type(self):
        m = ConcreteModel()

        initializer = ModularInitializerBase()
        initializer.submodel_initializers[ConcreteModel] = DummyInit

        assert isinstance(initializer.get_submodel_initializer(m), DummyInit)

    @pytest.mark.unit
    def test_get_submodel_initializer_model_default(self):
        m = ConcreteModel()
        m.default_initializer = DummyInit

        initializer = ModularInitializerBase()

        assert isinstance(initializer.get_submodel_initializer(m), DummyInit)

    @pytest.mark.unit
    def test_get_submodel_initializer_global_default(self):
        m = ConcreteModel()

        initializer = ModularInitializerBase(default_submodel_initializer=DummyInit)
        assert initializer.config.default_submodel_initializer == DummyInit

        assert isinstance(initializer.get_submodel_initializer(m), DummyInit)

    @pytest.mark.unit
    def test_get_submodel_initializer_model_w_params(self):
        class DummyParam:
            pass

        dummy_param = DummyParam()

        m = ConcreteModel()
        m.params = dummy_param

        initializer = ModularInitializerBase()
        initializer.submodel_initializers[dummy_param] = DummyInit

        assert isinstance(initializer.get_submodel_initializer(m), DummyInit)

    @pytest.mark.unit
    def test_get_submodel_initializer_none(self, caplog):
        m = ConcreteModel()

        initializer = ModularInitializerBase()

        assert initializer.get_submodel_initializer(m) is None
        expected = "No Initializer found for submodel unknown - attempting to continue."
        assert expected in caplog.text

    @pytest.mark.unit
    def test_get_submodel_initializer_process_base_default(self):
        m = ConcreteModel()
        m.b = ProcessBaseBlock()

        initializer = ModularInitializerBase()
        assert isinstance(
            initializer.get_submodel_initializer(m.b), BlockTriangularizationInitializer
        )

    @pytest.mark.unit
    def test_get_submodel_initializer_priority(self):
        # Progressively add higher priority initializers and ensure they are returned
        class DummyParam:
            def __init__(self):
                self.name = "dummy"

        dummy_param = DummyParam()

        m = ConcreteModel()
        m.params = dummy_param

        initializer = ModularInitializerBase()

        # No default set
        assert initializer.get_submodel_initializer(m) is None

        # Add a global default
        initializer.config.default_submodel_initializer = "global"
        assert initializer.get_submodel_initializer(m) == "global"

        # Parameter block
        initializer.submodel_initializers[dummy_param] = "params"
        assert initializer.get_submodel_initializer(m) == "params"

        # Model default
        m.default_initializer = "model_default"
        assert initializer.get_submodel_initializer(m) == "model_default"

        # Model type
        initializer.submodel_initializers[ConcreteModel] = "type"
        assert initializer.get_submodel_initializer(m) == "type"

        # Specific model
        initializer.submodel_initializers[m] = "specific_model"
        assert initializer.get_submodel_initializer(m) == "specific_model"

    @pytest.mark.unit
    def test_add_submodel_initializer(self):
        initializer = ModularInitializerBase()
        assert initializer.submodel_initializers == {}

        initializer.add_submodel_initializer("foo", "bar")
        assert initializer.submodel_initializers == {"foo": "bar"}

    @pytest.mark.unit
    def test_prepare_plugins_none(self):
        m = ConcreteModel()
        m.initialization_order = [m]

        initializer = ModularInitializerBase()

        args = {"foo": "bar"}
        subinit, plugin_args = initializer.prepare_plugins(m, args)

        assert subinit == {}
        assert plugin_args == {"foo": "bar"}
        # Make sure we didn't change the original args
        assert args == {"foo": "bar"}

    @pytest.mark.unit
    def test_prepare_plugins_no_args(self):
        m = ConcreteModel()
        m.b = Block()
        m.initialization_order = [m, m.b]

        initializer = ModularInitializerBase()
        initializer.add_submodel_initializer(m.b, DummyInit)

        args = {"foo": "bar"}
        subinit, plugin_args = initializer.prepare_plugins(m, args)

        assert len(subinit) == 1
        assert isinstance(subinit[m.b], DummyInit)
        assert plugin_args == {"foo": "bar", m.b: {}}
        assert m.b._test == "default"
        # Make sure we didn't change the original args
        assert args == {"foo": "bar"}

    @pytest.mark.unit
    def test_prepare_plugins_w_args(self):
        m = ConcreteModel()
        m.b = Block()
        m.initialization_order = [m, m.b]

        initializer = ModularInitializerBase()
        initializer.add_submodel_initializer(m.b, DummyInit)

        args = {"foo": "bar", m.b: {"output": "checkval"}}
        subinit, plugin_args = initializer.prepare_plugins(m, args)

        assert len(subinit) == 1
        assert isinstance(subinit[m.b], DummyInit)
        assert plugin_args == {"foo": "bar", m.b: {"output": "checkval"}}
        assert m.b._test == "checkval"
        # Make sure we didn't change the original args
        assert args == {"foo": "bar", m.b: {"output": "checkval"}}

    @pytest.mark.unit
    def test_solve_full_model_no_plugins(self):
        m = ConcreteModel()
        m.initialization_order = [m]

        initializer = ModularInitializerBase()

        results = initializer.solve_full_model(m, "test_results")

        assert results == "test_results"

    @pytest.mark.unit
    def test_solve_full_model_w_plugins(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.v = Var()
        m.b.c = Constraint(expr=m.b.v == 12)

        m.initialization_order = [m, m.b]

        initializer = ModularInitializerBase()

        results = initializer.solve_full_model(m, "test_results")

        assert value(m.b.v) == pytest.approx(12, rel=1e-8)
        assert isinstance(results, dict)
        assert "Solution" in results

    @pytest.mark.unit
    def test_cleanup_none(self):
        m = ConcreteModel()
        m.initialization_order = [m]

        initializer = ModularInitializerBase()

        args = {"foo": "bar"}
        subinit = {}
        initializer.cleanup(m, args, subinit)

        # No-op - what to assert?

    @pytest.mark.unit
    def test_cleanup_no_args(self):
        m = ConcreteModel()
        m.b = Block()
        m.initialization_order = [m, m.b]

        initializer = ModularInitializerBase()

        args = {"foo": "bar", m.b: {}}
        subinit = {m.b: DummyInit()}
        initializer.cleanup(m, args, subinit)

        assert m.b._test2 == "default"

    @pytest.mark.unit
    def test_cleanup_w_args(self):
        m = ConcreteModel()
        m.b = Block()
        m.initialization_order = [m, m.b]

        initializer = ModularInitializerBase()

        args = {"foo": "bar", m.b: {"output": "forty-two"}}
        subinit = {m.b: DummyInit()}
        initializer.cleanup(m, args, subinit)

        assert m.b._test2 == "forty-two"



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.network.arc import Arc

from pyomo.contrib.incidence_analysis import (
    IncidenceGraphInterface,
)
from pyomo.util.check_units import assert_units_consistent

import idaes.core as idaes
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models_extra.gas_distribution.properties.natural_gas import (
    NaturalGasParameterBlock,
)
from idaes.models_extra.gas_distribution.unit_models.pipeline import GasPipeline
from idaes.models_extra.gas_distribution.unit_models.node import PipelineNode
from idaes.models_extra.gas_distribution.unit_models.compressor import (
    IsothermalCompressor,
)

"""
These tests test construction of some slightly less trivial flowsheets
involving nodes, compressors, and pipelines.

These are construction tests only. We construct the flowsheet model,
fix degrees of freedom, and assert that the model is structurally
nonsingular. Tests that actually solve flowsheet models are included
in the gas_distribution/flowsheets/tests/ directory.

"""


@pytest.mark.unit
class TestConstructFlowsheets(unittest.TestCase):
    """ """

    def test_four_nodes_linear(self):
        """
        The network looks something like this:
        s            s
        |            |
        0 (c)-> 1 -> 2 (c)-> 3
                |            |
                d            d

        """
        m = pyo.ConcreteModel()
        fs_config = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**fs_config)
        m.fs.properties = NaturalGasParameterBlock()

        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline_set = pyo.Set(initialize=range(3))
        m.fs.pipeline = GasPipeline(m.fs.pipeline_set, **pipeline_config)

        node_config = [
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 0,
                "n_outlet_pipelines": 1,
                "n_supplies": 1,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 1,
                "n_supplies": 0,
                "n_demands": 1,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 1,
                "n_supplies": 1,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 0,
                "n_supplies": 0,
                "n_demands": 1,
            },
        ]
        node_config = {i: config for i, config in enumerate(node_config)}
        m.fs.node_set = pyo.Set(initialize=range(4))
        m.fs.node = PipelineNode(m.fs.node_set, initialize=node_config)

        compressor_config = {"property_package": m.fs.properties}
        m.fs.compressor_set = pyo.Set(initialize=range(2))
        m.fs.compressor = IsothermalCompressor(m.fs.compressor_set, **compressor_config)

        # Connect compressors to pipelines
        # Should/could I make this easier?
        pipeline_idx_map = {0: 0, 1: 2}

        def compressor_to_pipeline_rule(fs, i):
            return (
                m.fs.compressor[i].outlet_port,
                m.fs.pipeline[pipeline_idx_map[i]].inlet_port,
            )

        m.fs.compressor_to_pipeline = Arc(
            m.fs.compressor_set, rule=compressor_to_pipeline_rule
        )

        # Note that we are adding a compressor, not a pipeline, to the
        # outlet here.
        m.fs.node[0].add_pipeline_to_outlet(m.fs.compressor[0])
        m.fs.node[1].add_pipeline_to_inlet(m.fs.pipeline[0])
        m.fs.node[1].add_pipeline_to_outlet(m.fs.pipeline[1])
        m.fs.node[2].add_pipeline_to_inlet(m.fs.pipeline[1])
        m.fs.node[2].add_pipeline_to_outlet(m.fs.compressor[1])
        m.fs.node[3].add_pipeline_to_inlet(m.fs.pipeline[2])

        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        assert_units_consistent(m)

        ntfe = 2
        disc = pyo.TransformationFactory("dae.finite_difference")
        disc.apply_to(m, wrt=m.fs.time, nfe=ntfe)

        # Fix "design" variables
        for pipeline in m.fs.pipeline.values():
            pipeline.diameter.fix()
            pipeline.control_volume.length.fix()

        # Dynamic inputs:
        pred_dof = 10 * len(m.fs.time)
        # Initial conditions:
        pred_dof += 2 * (len(m.fs.pipeline[0].control_volume.length_domain) - 1)
        pred_dof += 2 * (len(m.fs.pipeline[1].control_volume.length_domain) - 1)
        pred_dof += 2 * (len(m.fs.pipeline[2].control_volume.length_domain) - 1)

        # Make sure we have the expected number of degrees of freedom
        self.assertEqual(degrees_of_freedom(m), pred_dof)

        # Fix predicted degrees of freedom:
        for t in m.fs.time:
            m.fs.node[0].supplies[0].state[t].mole_frac_comp[:].fix()
            m.fs.node[0].state[t].temperature.fix()
            m.fs.node[0].state[t].pressure.fix()
            m.fs.node[2].supplies[0].state[t].mole_frac_comp[:].fix()
            m.fs.node[2].supplies[0].state[t].flow_mol.fix()
            m.fs.node[2].supplies[0].state[t].temperature.fix()
            m.fs.node[1].demands[0].flow_mol[t].fix()
            m.fs.node[3].demands[0].flow_mol[t].fix()
            m.fs.compressor[0].boost_pressure[t].fix()
            m.fs.compressor[1].boost_pressure[t].fix()

        t0 = m.fs.time.first()
        x0 = m.fs.pipeline[0].control_volume.length_domain.first()
        xf = m.fs.pipeline[0].control_volume.length_domain.last()
        for x in m.fs.pipeline[0].control_volume.length_domain:
            # Here I assume that all three pipelines have the same
            # length domain.
            if x != x0:
                m.fs.pipeline[0].control_volume.pressure[t0, x].fix()
                m.fs.pipeline[1].control_volume.pressure[t0, x].fix()
                m.fs.pipeline[2].control_volume.pressure[t0, x].fix()
            if x != xf:
                m.fs.pipeline[0].control_volume.flow_mass[t0, x].fix()
                m.fs.pipeline[1].control_volume.flow_mass[t0, x].fix()
                m.fs.pipeline[2].control_volume.flow_mass[t0, x].fix()

        # Make sure we have zero degrees of freedom and a perfect matching
        # between constraints and variables.
        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        N, M = len(igraph.constraints), len(igraph.variables)
        matching = igraph.maximum_matching()
        self.assertEqual(len(matching), N)

    def test_two_supply_one_demand(self):
        """
        The network looks something like this:
        s-0 
           \
            --> 2 (c)-> 3-d
           /
        s-1
        """
        m = pyo.ConcreteModel()
        fs_config = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**fs_config)
        m.fs.properties = NaturalGasParameterBlock()

        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline_set = pyo.Set(initialize=range(3))
        m.fs.pipeline = GasPipeline(m.fs.pipeline_set, **pipeline_config)

        node_config = [
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 0,
                "n_outlet_pipelines": 1,
                "n_supplies": 1,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 0,
                "n_outlet_pipelines": 1,
                "n_supplies": 1,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 2,
                "n_outlet_pipelines": 1,
                "n_supplies": 0,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 0,
                "n_supplies": 0,
                "n_demands": 1,
            },
        ]
        node_config = {i: config for i, config in enumerate(node_config)}
        m.fs.node_set = pyo.Set(initialize=range(4))
        m.fs.node = PipelineNode(m.fs.node_set, initialize=node_config)

        compressor_config = {"property_package": m.fs.properties}
        m.fs.compressor_set = pyo.Set(initialize=range(1))
        m.fs.compressor = IsothermalCompressor(m.fs.compressor_set, **compressor_config)

        # Connect compressors to pipelines
        # Should/could I make this easier?
        pipeline_idx_map = {0: 2}

        def compressor_to_pipeline_rule(fs, i):
            return (
                m.fs.compressor[i].outlet_port,
                m.fs.pipeline[pipeline_idx_map[i]].inlet_port,
            )

        m.fs.compressor_to_pipeline = Arc(
            m.fs.compressor_set, rule=compressor_to_pipeline_rule
        )

        m.fs.node[0].add_pipeline_to_outlet(m.fs.pipeline[0])
        m.fs.node[1].add_pipeline_to_outlet(m.fs.pipeline[1])
        m.fs.node[2].add_pipeline_to_inlet(m.fs.pipeline[0])
        m.fs.node[2].add_pipeline_to_inlet(m.fs.pipeline[1])
        m.fs.node[2].add_pipeline_to_outlet(m.fs.compressor[0])
        m.fs.node[3].add_pipeline_to_inlet(m.fs.pipeline[2])

        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        assert_units_consistent(m)

        ntfe = 2
        disc = pyo.TransformationFactory("dae.finite_difference")
        disc.apply_to(m, wrt=m.fs.time, nfe=ntfe)

        # Fix "design" variables
        for pipeline in m.fs.pipeline.values():
            pipeline.diameter.fix()
            pipeline.control_volume.length.fix()

        # Predicted degrees of freedom:
        pred_dof = 8 * len(m.fs.time)
        pred_dof += 2 * (len(m.fs.pipeline[0].control_volume.length_domain) - 1)
        pred_dof += 2 * (len(m.fs.pipeline[1].control_volume.length_domain) - 1)
        pred_dof += 2 * (len(m.fs.pipeline[2].control_volume.length_domain) - 1)

        self.assertEqual(degrees_of_freedom(m), pred_dof)

        # Fix predicted degrees of freedom:
        # Dynamic inputs:
        for t in m.fs.time:
            m.fs.node[0].supplies[0].state[t].mole_frac_comp[:].fix()
            m.fs.node[0].state[t].temperature.fix()
            m.fs.node[0].state[t].pressure.fix()
            m.fs.node[1].supplies[0].state[t].mole_frac_comp[:].fix()
            m.fs.node[1].supplies[0].state[t].flow_mol.fix()
            m.fs.node[1].supplies[0].state[t].temperature.fix()
            m.fs.node[3].demands[0].flow_mol[t].fix()
            m.fs.compressor[0].boost_pressure[t].fix()

        # Initial conditions:
        t0 = m.fs.time.first()
        x0 = m.fs.pipeline[0].control_volume.length_domain.first()
        xf = m.fs.pipeline[0].control_volume.length_domain.last()
        for x in m.fs.pipeline[0].control_volume.length_domain:
            # Here I assume that all three pipelines have the same
            # length domain.
            if x != x0:
                m.fs.pipeline[0].control_volume.pressure[t0, x].fix()
                m.fs.pipeline[1].control_volume.flow_mass[t0, x].fix()
                m.fs.pipeline[2].control_volume.pressure[t0, x].fix()
            if x != xf:
                m.fs.pipeline[0].control_volume.flow_mass[t0, x].fix()
                m.fs.pipeline[1].control_volume.pressure[t0, x].fix()
                m.fs.pipeline[2].control_volume.flow_mass[t0, x].fix()

        # Make sure we have zero degrees of freedom and a perfect matching
        # between constraints and variables.
        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        N, M = len(igraph.constraints), len(igraph.variables)
        matching = igraph.maximum_matching()
        self.assertEqual(len(matching), N)

    def test_one_supply_two_demand(self):
        """
        The network looks something like this:
                   (c)-> 2-d
                  /
        s-0 -> 1 - 
                  \
                   (c)-> 3-d
        """
        m = pyo.ConcreteModel()
        fs_config = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**fs_config)
        m.fs.properties = NaturalGasParameterBlock()

        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline_set = pyo.Set(initialize=range(3))
        m.fs.pipeline = GasPipeline(m.fs.pipeline_set, **pipeline_config)

        node_config = [
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 0,
                "n_outlet_pipelines": 1,
                "n_supplies": 1,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 2,
                "n_supplies": 0,
                "n_demands": 0,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 0,
                "n_supplies": 0,
                "n_demands": 1,
            },
            {
                "property_package": m.fs.properties,
                "n_inlet_pipelines": 1,
                "n_outlet_pipelines": 0,
                "n_supplies": 0,
                "n_demands": 1,
            },
        ]
        node_config = {i: config for i, config in enumerate(node_config)}
        m.fs.node_set = pyo.Set(initialize=range(4))
        m.fs.node = PipelineNode(m.fs.node_set, initialize=node_config)

        compressor_config = {"property_package": m.fs.properties}
        m.fs.compressor_set = pyo.Set(initialize=range(2))
        m.fs.compressor = IsothermalCompressor(m.fs.compressor_set, **compressor_config)

        # Connect compressors to pipelines
        # Should/could I make this easier?
        pipeline_idx_map = {0: 1, 1: 2}

        def compressor_to_pipeline_rule(fs, i):
            return (
                m.fs.compressor[i].outlet_port,
                m.fs.pipeline[pipeline_idx_map[i]].inlet_port,
            )

        m.fs.compressor_to_pipeline = Arc(
            m.fs.compressor_set, rule=compressor_to_pipeline_rule
        )

        m.fs.node[0].add_pipeline_to_outlet(m.fs.pipeline[0])
        m.fs.node[1].add_pipeline_to_inlet(m.fs.pipeline[0])
        m.fs.node[1].add_pipeline_to_outlet(m.fs.compressor[0])
        m.fs.node[1].add_pipeline_to_outlet(m.fs.compressor[1])
        m.fs.node[2].add_pipeline_to_inlet(m.fs.pipeline[1])
        m.fs.node[3].add_pipeline_to_inlet(m.fs.pipeline[2])

        expand_arcs = pyo.TransformationFactory("network.expand_arcs")
        expand_arcs.apply_to(m)

        assert_units_consistent(m)

        ntfe = 2
        disc = pyo.TransformationFactory("dae.finite_difference")
        disc.apply_to(m, wrt=m.fs.time, nfe=ntfe)

        # Fix "design" variables
        for pipeline in m.fs.pipeline.values():
            pipeline.diameter.fix()
            pipeline.control_volume.length.fix()

        # Predicted degrees of freedom (assuming all pipelines have same length
        # discretization)
        pred_dof = 7 * len(m.fs.time)
        pred_dof += 6 * (len(m.fs.pipeline[0].control_volume.length_domain) - 1)
        self.assertEqual(degrees_of_freedom(m), pred_dof)

        # Fix predicted degrees of freedom:
        # Dynamic inputs:
        for t in m.fs.time:
            m.fs.node[0].supplies[0].state[t].mole_frac_comp[:].fix()
            m.fs.node[0].state[t].temperature.fix()
            m.fs.node[0].state[t].pressure.fix()
            m.fs.compressor[0].boost_pressure[t].fix()
            m.fs.compressor[1].boost_pressure[t].fix()
            m.fs.node[2].demands[0].flow_mol[t].fix()
            m.fs.node[3].demands[0].flow_mol[t].fix()

        # Initial conditions:
        t0 = m.fs.time.first()
        x0 = m.fs.pipeline[0].control_volume.length_domain.first()
        xf = m.fs.pipeline[0].control_volume.length_domain.last()
        for x in m.fs.pipeline[0].control_volume.length_domain:
            # Here I assume that all three pipelines have the same
            # length domain.
            if x != x0:
                m.fs.pipeline[0].control_volume.pressure[t0, x].fix()
                m.fs.pipeline[1].control_volume.pressure[t0, x].fix()
                m.fs.pipeline[2].control_volume.pressure[t0, x].fix()
            if x != xf:
                m.fs.pipeline[0].control_volume.flow_mass[t0, x].fix()
                m.fs.pipeline[1].control_volume.flow_mass[t0, x].fix()
                m.fs.pipeline[2].control_volume.flow_mass[t0, x].fix()

        # Make sure we have zero degrees of freedom and a perfect matching
        # between constraints and variables.
        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        N, M = len(igraph.constraints), len(igraph.variables)
        matching = igraph.maximum_matching()
        self.assertEqual(len(matching), N)


if __name__ == "__main__":
    unittest.main()
